@@ -1,4 +1,4 @@
-import time
+import uuid
 import json
 import requests
 
@@ -13,7 +13,10 @@ from todoist.managers.project_notes import ProjectNotesManager
 from todoist.managers.items import ItemsManager
 from todoist.managers.labels import LabelsManager
 from todoist.managers.reminders import RemindersManager
+from todoist.managers.locations import LocationsManager
 from todoist.managers.user import UserManager
+from todoist.managers.collaborators import CollaboratorsManager
+from todoist.managers.collaborator_states import CollaboratorStatesManager
 
 
 class TodoistAPI(object):
@@ -21,8 +24,8 @@ class TodoistAPI(object):
     Implements the API that makes it possible to interact with a Todoist user
     account and its data.
     """
-    _serialize_fields = ('token', 'api_url', 'seq_no', 'seq_no_partial',
-                         'state', 'temp_ids', 'timestamp', 'timestamp_suffix')
+    _serialize_fields = ('token', 'api_endpoint', 'seq_no', 'seq_no_partial',
+                         'state', 'temp_ids')
 
     @classmethod
     def deserialize(cls, data):
@@ -33,7 +36,7 @@ class TodoistAPI(object):
         return obj
 
     def __init__(self, token='', api_endpoint='https://api.todoist.com'):
-        self.api_url = '%s/API/v6/' % api_endpoint  # Todoist API
+        self.api_endpoint = api_endpoint
         self.seq_no = 0  # Sequence number since last update
         self.seq_no_partial = {}  # Sequence number of partial syncs
         self.state = {  # Local copy of all of the user's objects
@@ -46,6 +49,7 @@ class TodoistAPI(object):
             'Labels': [],
             'LiveNotifications': [],
             'LiveNotificationsLastRead': -1,
+            'Locations': [],
             'Notes': [],
             'ProjectNotes': [],
             'Projects': [],
@@ -59,8 +63,6 @@ class TodoistAPI(object):
         self.token = token  # User's API token
         self.temp_ids = {}  # Mapping of temporary ids to real ids
         self.queue = []  # Requests to be sent are appended here
-        self.timestamp = -1
-        self.timestamp_suffix = -1
 
         # managers
         self.projects = ProjectsManager(self)
@@ -71,9 +73,12 @@ class TodoistAPI(object):
         self.notes = NotesManager(self)
         self.live_notifications = LiveNotificationsManager(self)
         self.reminders = RemindersManager(self)
+        self.locations = LocationsManager(self)
         self.invitations = InvitationsManager(self)
         self.biz_invitations = BizInvitationsManager(self)
         self.user = UserManager(self)
+        self.collaborators = CollaboratorsManager(self)
+        self.collaborator_states = CollaboratorStatesManager(self)
 
     def __getitem__(self, key):
         return self.state[key]
@@ -81,6 +86,9 @@ class TodoistAPI(object):
     def serialize(self):
         return {key: getattr(self, key) for key in self._serialize_fields}
 
+    def get_api_url(self):
+        return '%s/API/v6/' % self.api_endpoint
+
     def _update_state(self, syncdata):
         """
         Updates the local state, with the data returned by the server after a
@@ -100,6 +108,8 @@ class TodoistAPI(object):
         if 'LiveNotificationsLastRead' in syncdata:
             self.state['LiveNotificationsLastRead'] = \
                 syncdata['LiveNotificationsLastRead']
+        if 'Locations' in syncdata:
+            self.state['Locations'] = syncdata['Locations']
         if 'Settings' in syncdata:
             self.state['Settings'].update(syncdata['Settings'])
         if 'SettingsNotifications' in syncdata:
@@ -150,10 +160,10 @@ class TodoistAPI(object):
         returned, and if not, then None is returned.
         """
         if objtype == 'Collaborators':
-            return self.collaborators_get_by_id(obj['id'])
+            return self.collaborators.get_by_id(obj['id'])
         elif objtype == 'CollaboratorStates':
-            return self.collaborator_state_get_by_ids(obj['project_id'],
-                                                      obj['user_id'])
+            return self.collaborator_states.get_by_ids(obj['project_id'],
+                                                       obj['user_id'])
         elif objtype == 'Filters':
             return self.filters.get_by_id(obj['id'], only_local=True)
         elif objtype == 'Items':
@@ -178,27 +188,32 @@ class TodoistAPI(object):
         Calculates what is the seq_no that should be sent, based on the last
         seq_no and the resource_types that are requested.
         """
-        seq_no = self.seq_no
+        seq_no = -1
         if resource_types:
             for resource in resource_types:
-                if resource in self.seq_no_partial:
-                    if seq_no == 0 or self.seq_no_partial[resource] < seq_no:
+                if resource not in self.seq_no_partial:
+                    seq_no = self.seq_no
+                else:
+                    if seq_no == -1 or self.seq_no_partial[resource] < seq_no:
                         seq_no = self.seq_no_partial[resource]
-        return seq_no
+        if seq_no == -1:
+            return self.seq_no
+        else:
+            return seq_no
 
     def _update_seq_no(self, seq_no, resource_types):
         """
         Updates the seq_no and the seq_no_partial, based on the seq_no in
         the response and the resource_types that were requested.
         """
-        if not seq_no:
-            return
-        if resource_types and seq_no > self.seq_no:
-            for resource in resource_types:
-                self.seq_no_partial[resource] = seq_no
-        else:
+        if not seq_no or not resource_types:
+            pass
+        elif 'all' in resource_types:
             self.seq_no = seq_no
             self.seq_no_partial = {}
+        elif resource_types and seq_no > self.seq_no:
+            for resource in resource_types:
+                self.seq_no_partial[resource] = seq_no
 
     def _replace_temp_id(self, temp_id, new_id):
         """
@@ -222,7 +237,7 @@ class TodoistAPI(object):
         object received (if any), or whatever answer it got otherwise.
         """
         if not url:
-            url = self.api_url
+            url = self.get_api_url()
         response = requests.get(url + call, **kwargs)
         try:
             return response.json()
@@ -235,7 +250,7 @@ class TodoistAPI(object):
         object received (if any), or whatever answer it got otherwise.
         """
         if not url:
-            url = self.api_url
+            url = self.get_api_url()
         response = requests.post(url + call, **kwargs)
         try:
             return response.json()
@@ -243,17 +258,11 @@ class TodoistAPI(object):
             return response.text
 
     # Sync
-    def generate_timestamp(self):
+    def generate_uuid(self):
         """
-        Generates a timestamp, which is based on the current unix time.
+        Generates a uuid.
         """
-        now = int(time.time())
-        if now != self.timestamp:
-            self.timestamp = now
-            self.timestamp_suffix = 1
-        else:
-            self.timestamp_suffix += 1
-        return str(self.timestamp) + '.' + str(self.timestamp_suffix)
+        return str(uuid.uuid1())
 
     def sync(self, commands=None, **kwargs):
         """
@@ -261,19 +270,22 @@ class TodoistAPI(object):
         fetches the latest updated data from the server.
         """
         params = {
-            'seq_no': self._get_seq_no(kwargs.get('resource_types', None)),
             'token': self.token,
             'commands': json.dumps(commands or []),
             'day_orders_timestamp': self.state['DayOrdersTimestamp'],
         }
+        if not commands:
+            params['seq_no'] = self._get_seq_no(kwargs.get('resource_types',
+                                                           None)),
         if 'include_notification_settings' in kwargs:
             params['include_notification_settings'] = 1
         if 'resource_types' in kwargs:
             params['resource_types'] = json.dumps(kwargs['resource_types'])
         data = self._post('sync', params=params)
         self._update_state(data)
-        self._update_seq_no(data.get('seq_no', None),
-                            kwargs.get('resource_types', None))
+        if not commands:
+            self._update_seq_no(data.get('seq_no', None),
+                                kwargs.get('resource_types', None))
 
         return data
 
@@ -292,7 +304,9 @@ class TodoistAPI(object):
             for temp_id, new_id in ret['TempIdMapping'].items():
                 self.temp_ids[temp_id] = new_id
                 self._replace_temp_id(temp_id, new_id)
-        return ret['SyncStatus']
+        if 'SyncStatus' in ret:
+            return ret['SyncStatus']
+        return ret
 
     # Authentication
     def login(self, email, password):
@@ -330,15 +344,6 @@ class TodoistAPI(object):
             self.token = data['token']
         return data
 
-    def delete_user(self, current_password, **kwargs):
-        """
-        Deletes an existing user.
-        """
-        params = {'token': self.token,
-                  'current_password': current_password}
-        params.update(kwargs)
-        return self._get('delete_user', params=params)
-
     # Miscellaneous
     def upload_file(self, filename, **kwargs):
         """
@@ -347,7 +352,7 @@ class TodoistAPI(object):
         params = {'token': self.token}
         params.update(kwargs)
         files = {'file': open(filename, 'rb')}
-        return self._post('upload_file', self.api_url, params=params,
+        return self._post('upload_file', self.get_api_url(), params=params,
                           files=files)
 
     def query(self, queries, **kwargs):
@@ -406,11 +411,10 @@ class TodoistAPI(object):
         """
         Appends a request to the queue, to share a project with a user.
         """
-        ts = self.generate_timestamp()
         item = {
             'type': 'share_project',
-            'temp_id': '$' + ts,
-            'timestamp': ts,
+            'temp_id': self.generate_uuid(),
+            'uuid': self.generate_uuid(),
             'args': {
                 'project_id': project_id,
                 'email': email,
@@ -426,7 +430,7 @@ class TodoistAPI(object):
         """
         item = {
             'type': 'delete_collaborator',
-            'timestamp': self.generate_timestamp(),
+            'uuid': self.generate_uuid(),
             'args': {
                 'project_id': project_id,
                 'email': email,
@@ -440,32 +444,13 @@ class TodoistAPI(object):
         """
         item = {
             'type': 'take_ownership',
-            'timestamp': self.generate_timestamp(),
+            'uuid': self.generate_uuid(),
             'args': {
                 'project_id': project_id,
             },
         }
         self.queue.append(item)
 
-    def collaborators_get_by_id(self, user_id):
-        """
-        Finds and returns the collaborator based on the user id.
-        """
-        for obj in self.state['Collaborators']:
-            if obj['id'] == user_id:
-                return obj
-        return None
-
-    def collaborator_state_get_by_ids(self, project_id, user_id):
-        """
-        Finds and returns the collaborator state based on the project and user
-        id.
-        """
-        for obj in self.state['CollaboratorStates']:
-            if obj['project_id'] == project_id and obj['user_id'] == user_id:
-                return obj
-        return None
-
     # Auxiliary
     def get_project(self, project_id):
         """
@@ -473,7 +458,13 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'project_id': project_id}
-        return self._get('get_project', params=params)
+        data = self._get('get_project', params=params)
+        obj = data.get('project', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Projects': [obj]})
+            return [o for o in self.state['Projects']
+                    if o['id'] == obj['id']][0]
+        return None
 
     def get_item(self, item_id):
         """
@@ -481,7 +472,12 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'item_id': item_id}
-        return self._get('get_item', params=params)
+        data = self._get('get_item', params=params)
+        obj = data.get('item', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Items': [obj]})
+            return [o for o in self.state['Items'] if o['id'] == obj['id']][0]
+        return None
 
     def get_label(self, label_id):
         """
@@ -489,7 +485,12 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'label_id': label_id}
-        return self._get('get_label', params=params)
+        data = self._get('get_label', params=params)
+        obj = data.get('label', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Labels': [obj]})
+            return [o for o in self.state['Labels'] if o['id'] == obj['id']][0]
+        return None
 
     def get_note(self, note_id):
         """
@@ -497,7 +498,12 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'note_id': note_id}
-        return self._get('get_note', params=params)
+        data = self._get('get_note', params=params)
+        obj = data.get('note', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Notes': [obj]})
+            return [o for o in self.state['Notes'] if o['id'] == obj['id']][0]
+        return None
 
     def get_filter(self, filter_id):
         """
@@ -505,7 +511,13 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'filter_id': filter_id}
-        return self._get('get_filter', params=params)
+        data = self._get('get_filter', params=params)
+        obj = data.get('filter', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Filters': [obj]})
+            return [o for o in self.state['Filters']
+                    if o['id'] == obj['id']][0]
+        return None
 
     def get_reminder(self, reminder_id):
         """
@@ -513,7 +525,13 @@ class TodoistAPI(object):
         """
         params = {'token': self.token,
                   'reminder_id': reminder_id}
-        return self._get('get_reminder', params=params)
+        data = self._get('get_reminder', params=params)
+        obj = data.get('reminder', None)
+        if obj and 'error' not in obj:
+            self._update_state({'Reminders': [obj]})
+            return [o for o in self.state['Reminders']
+                    if o['id'] == obj['id']][0]
+        return None
 
     # Class
     def __repr__(self):
