@@ -1,12 +1,13 @@
 # -*- coding: utf-8 -*-
 from .. import models
-from .generic import Manager, AllMixin, GetByIdMixin
+from .generic import Manager, AllMixin, GetByIdMixin, SyncMixin
 
 
-class LabelsManager(Manager, AllMixin, GetByIdMixin):
+class LabelsManager(Manager, AllMixin, GetByIdMixin, SyncMixin):
 
     state_name = 'Labels'
     object_type = 'label'
+    resource_type = 'labels'
 
     def register(self, name, **kwargs):
         """
@@ -14,14 +15,13 @@ class LabelsManager(Manager, AllMixin, GetByIdMixin):
         request to the queue.
         """
         obj = models.Label({'name': name}, self.api)
-        ts = self.api.generate_timestamp()
-        obj.temp_id = obj['id'] = '$' + ts
+        obj.temp_id = obj['id'] = self.api.generate_uuid()
         obj.data.update(kwargs)
         self.state[self.state_name].append(obj)
         item = {
             'type': 'label_register',
             'temp_id': obj.temp_id,
-            'timestamp': ts,
+            'uuid': self.api.generate_uuid(),
             'args': obj.data,
         }
         self.queue.append(item)
